define('feature/comments/anchors', ['jquery', 'model/page-state'], function ($, pageState) {




    /**
     * Details for anchoring to a pull-requests general comments
     * @param opt_pullRequestInfo {Object} containing projectKey, repoSlug, and pullRequestId properties
     * @api public
     * @constructor
     */
    function PullRequestAnchor(opt_pullRequestInfo) {
        opt_pullRequestInfo = opt_pullRequestInfo || {};
        this._projectKey = opt_pullRequestInfo.projectKey || pageState.getProject().key;
        this._repoSlug = opt_pullRequestInfo.repoSlug || pageState.getRepository().slug;
        this._pullRequestId = opt_pullRequestInfo.pullRequestId || pageState.getPullRequest().getId();
    }
    PullRequestAnchor.prototype.getId = function() {
        return 'pullrequest-' + this._projectKey + '-' + this._repoSlug + '-' + this._pullRequestId;
    };
    PullRequestAnchor.prototype.toJSON = function() {
        // the server doesn't want any anchor info. It NPE's if we provide it.
        return undefined;
    };



    /**
     * Details for anchoring to a diff
     * @param fileChange {FileChange} a change object for this diff.
     * @api public
     * @constructor
     */
    function DiffAnchor(fileChange) {
        this._path = fileChange.getPath();
        this._srcPath = fileChange.getSrcPath();
        this._commitRange = fileChange.getCommitRange();
        this._projectKey = pageState.getProject().key;
        this._repoSlug = pageState.getRepository().slug;
    }
    DiffAnchor.prototype.getId = function() {
        return 'diff-' + this._projectKey + '-' + this._repoSlug + '-' + this._commitRange.getId() + '-' + this._path;
    };
    DiffAnchor.prototype.toJSON = function() {
        var json = {
            path : this._path.toString(),
            commitRange : this._commitRange.toJSON()
        };
        var srcPath = this._srcPath && this._srcPath.toString();
        if (srcPath) { // backend can't handle an empty srcPath.
            json.srcPath = srcPath;
        }
        return json;
    };



    /**
     * Details for anchoring to a specific line in a diff
     * @param diffAnchor {DiffAnchor} parent diff anchor
     * @param lineType {String} line type - 'ADDED', 'REMOVED', 'CONTEXT'
     * @param lineNumber {Number} - source line number for 'REMOVED' or 'CONTEXT', destination line number for 'ADDED'
     * @api private
     * @constructor
     */
    function LineAnchor(diffAnchor, lineType, lineNumber) {
        this._diffAnchor = diffAnchor;
        this._lineType = lineType;
        this._line = lineNumber;
    }
    LineAnchor.prototype.getId = function() {
        return 'line-' + this._diffAnchor.getId() + '-' + this._lineType + '-' + this._line;
    };
    LineAnchor.prototype.toJSON = function() {
        return $.extend(this._diffAnchor.toJSON(), {
            line : this._line,
            lineType : this._lineType
        });
    };



    return {
        PullRequestAnchor : PullRequestAnchor,
        DiffAnchor : DiffAnchor,
        LineAnchor : LineAnchor
    };
});
