define('feature/pull-request/pull-request-activity', [
    'jquery',
    'eve',
    'underscore',
    'util/ajax',
    'util/navbuilder',
    'util/client-storage',
    'model/page-state',
    'model/pull-request',
    'model/commit-range',
    'model/file-change',
    'model/page-state',
    'widget/paged-scrollable',
    'feature/comments',
    'feature/file-content'
], function(
    $,
    eve,
    _,
    ajaxUtil,
    navbuilder,
    clientStorage,
    pageState,
    PullRequest,
    CommitRange,
    FileChange,
    PageState,
    PagedScrollable,
    comments,
    FileContent
) {

        function bindAddGeneralCommentHandler(keys) {
            (this.execute ? this : AJS.whenIType(keys)).execute(function() {
                var $form = $(".general-comment-form");
                $('html, body').animate({
                    scrollTop: Math.max(0, $form.offset().top - $(window).height() / 2)
                });
                $form.find("textarea").focus();
            });
        }

        function commentContainerDestroyedHandler() {
            this.$el.closest('.diff-comment-activity').remove();
        }

        function PullRequestActivity(contextSelector, pullRequest, fromType, fromId, options) {
            this._$container = $(contextSelector);
            this.pullRequest = pullRequest;
            this.pullRequestPathComponents = {
                projectKey : pageState.getProject().key,
                repoSlug : pageState.getRepository().slug,
                pullRequestId : this.pullRequest.getId()
            };
            this.fromType = fromType;
            this.fromId = fromId;

            this._$spinner = $('<div class="spinner"/>').insertAfter(this._$container);

            PagedScrollable.call(this, options.scrollableElement || contextSelector, {
                pageSize : 25, /* Makes a lot of git calls on each request. Fewer items => faster response. */
                dataLoadedEvent : 'stash.feature.pullRequestActivity.dataLoaded'
            });
        }
        $.extend(PullRequestActivity.prototype, PagedScrollable.prototype);

        PullRequestActivity.prototype.init = function(options) {
            this.renderedDiffFileContents = [];
            PagedScrollable.prototype.init.call(this, options);

            if (!this.loadedRange.isEmpty()) { // render the initial diffs now since attachNewContent won't be called.
                this.renderedDiffFileContents = this.renderedDiffFileContents.concat(
                    PullRequestActivity.renderDiffs(
                        this._$container.children('.diff-comment-activity'),
                        options.diffCommentData,
                        new CommitRange({
                            pullRequest : this.pullRequest
                        })
                    )
                );
            }

            comments.bindContext(this._$container, new comments.PullRequestAnchor(this.pullRequestPathComponents));
            var self = this;
            this._approvalHandler = function(data) {
                self.addApprovalActivity(data.approved, data.user, new Date());
            };
            eve.on('stash.widget.approve-button.added', this._approvalHandler);
            eve.on('stash.widget.approve-button.removed', this._approvalHandler);
            eve.on('stash.comment.comment-container.destroyed', commentContainerDestroyedHandler);
            eve.on('stash.keyboard.shortcuts.pullrequest.requestAddGeneralCommentHandler', bindAddGeneralCommentHandler);
        };

        PullRequestActivity.prototype.reset = function() {
            eve.off('stash.widget.approve-button.added', this._approvalHandler);
            eve.off('stash.widget.approve-button.removed', this._approvalHandler);
            eve.off('stash.comment.comment-container.destroyed', commentContainerDestroyedHandler);
            eve.off('stash.keyboard.shortcuts.pullrequest.requestAddGeneralCommentHandler', bindAddGeneralCommentHandler);

            comments.unbindContext(this._$container);

            _.invoke(this.renderedDiffFileContents, 'destroy');
            delete this.renderedDiffFileContents;

            PagedScrollable.prototype.reset.call(this);
            this.currentTime = undefined;
        };

        PullRequestActivity.prototype.checkCommentIsNew = function(comment) {
            comment.isUnread =
                comment.updatedDate > this.lastViewed &&
                    comment.author.name !== PageState.getCurrentUser().name;

            if (comment.comments.length) {
                _.each(comment.comments, _.bind(this.checkCommentIsNew, this));
            }
        };

        PullRequestActivity.prototype.checkCommentActivitiesAreNew = function(activities) {
            var self = this;
            _.each(activities, function (activity) {
                if (activity.action === 'COMMENTED') {
                    self.checkCommentIsNew(activity.comment);
                }
            });
        };

        PullRequestActivity.prototype.requestData = function(start, limit) {
            var self = this;
            //Use permalink params only for the first page request where the current page url is clearly a permalink
            var permalinkParams = start === 0 && !_.isUndefined(this.fromType) && !_.isUndefined(this.fromId) ?
                    { fromType: this.fromType, fromId: this.fromId } : {};

            this._$spinner.spin('large');
            return ajaxUtil.rest({
                url : navbuilder.rest()
                      .project(this.pullRequestPathComponents.projectKey)
                      .repo(this.pullRequestPathComponents.repoSlug)
                      .pullRequest(this.pullRequestPathComponents.pullRequestId)
                      .activities()
                      .withParams($.extend(permalinkParams, {
                          start : start,
                          limit : limit,
                          avatarSize: stash.widget.avatarSizeInPx({ size: 'medium' }),
                          markup: true
                      }))
                      .build(),
                statusCode : {
                    //A 404 should only ever happen if the current URL is for a permalink and the comment has been deleted
                    //or the comment or activity ID is invalid
                    '404' :  function(xhr, textStatus, errorThrown, errors, dominantError) {
                        var title =
                            self.fromType === 'activity' ? stash_i18n("stash.web.pull-request.activity.notfound.title", "Activity not found")
                                                         : stash_i18n("stash.web.pull-request.comment.notfound.title", "Comment not found");
                        var message =
                            self.fromType === 'activity' ? stash_i18n("stash.web.pull-request.activity.notfound.message", "The requested activity does not exist.")
                                                         : stash_i18n("stash.web.pull-request.comment.notfound.message", "The requested comment does not exist. It may have been deleted.");
                        return $.extend({}, dominantError, {
                            title: title,
                            titleClass: 'confirm-header',
                            message: message,
                            fallbackTitle: stash_i18n("stash.web.pull-request.activity.notfound.fallback.title", "Ok"),
                            fallbackUrl: navbuilder.project(self.pullRequestPathComponents.projectKey)
                                            .repo(self.pullRequestPathComponents.repoSlug)
                                            .pullRequest(self.pullRequestPathComponents.pullRequestId)
                                            .overview().build(),
                            canClose: false,
                            shouldReload: false
                        });
                    }
                }
            }).done(function(data, textStatus, xhr) {
                // Only set the currentTime on the first rest request for activity
                if (!self.currentTime) {
                    // Use server time otherwise fallback to client time
                    var currentTime = new Date(xhr.getResponseHeader('Date')).getTime(),
                        lastViewedKey = clientStorage.buildKey('last-viewed', 'pull-request');

                    self.currentTime = isNaN(currentTime) ? new Date().getTime() : currentTime;

                    // If first time viewing, we don't want all the comments to be marked as unread
                    self.lastViewed = clientStorage.getItem(lastViewedKey) || self.currentTime;
                    clientStorage.setItem(lastViewedKey, self.currentTime);
                }

                self.checkCommentActivitiesAreNew(data.values);
            }).fail(function() {
                self._$spinner.spinStop();
            });
        };

        PullRequestActivity.prototype.attachContent = function attach(method, elem) {
            this._$container[method === 'html' ? 'append' : method](elem);
        };

        PullRequestActivity.prototype.decorateForFocus = function(data) {
            var isActivityPermalink = this.fromType === 'activity';
            if (isActivityPermalink) {
                var activityId = parseInt(this.fromId, 10);
                _.each(data.values, function (activity) {
                    if (activity.id === activityId) {
                        activity.isFocused = true;
                    }
                });
            } else {
                var commentId = parseInt(this.fromId, 10);

                var focusComment = function (comment) {
                    if (comment.id === commentId) {
                        comment.isFocused = true;
                    } else if (comment.comments) {
                        _.each(comment.comments, function (reply) {
                            focusComment(reply);
                        });
                    }
                };

                _.each(data.values, function (activity) {
                    if (activity.comment) {
                        focusComment(activity.comment);
                    }
                });
            }
        };

        PullRequestActivity.prototype.onAttachFirstPermalinkPage = function(data, attachmentMethod) {
            var self = this;

            var $loadPrevious = $(stash.feature.pullRequest.loadPreviousActivities());
            this.attachContent(attachmentMethod, $loadPrevious);
            var $loadPreviousLink = $loadPrevious.find('a');

            var $topSpinner = $loadPrevious.append($('<div class="spinner"/>'));

            var fromId = data.previousPageStartId,
                lastFromId = data.values[0].id;

            function loadPreviousActivities() {
                $loadPreviousLink.hide();
                $topSpinner.spin('large');

                ajaxUtil.rest({
                    url : navbuilder.rest()
                        .project(self.pullRequestPathComponents.projectKey)
                        .repo(self.pullRequestPathComponents.repoSlug)
                        .pullRequest(self.pullRequestPathComponents.pullRequestId)
                        .activities()
                        .withParams({
                            start : 0,
                            limit : data.limit,
                            fromType: 'activity',
                            fromId: fromId,
                            avatarSize: stash.widget.avatarSizeInPx({ size: 'medium' }),
                            markup: true
                        })
                        .build()
                }).done(function(data, textStatus, xhr) {
                    //Remove duplicates from this new previous page - this may happen
                    //when loading the very first page because we've run out of activities
                    var oldActivityIndex;
                    _.any(data.values, function(activity, index) {
                        if (activity.id === lastFromId) {
                            oldActivityIndex = index;
                            return true;
                        }
                    });
                    var nonDuplicates = data.values.slice(0, oldActivityIndex);

                    self.checkCommentActivitiesAreNew(nonDuplicates);

                    //Attach new activities after the load-previous button
                    self.attachActivities(nonDuplicates, function(element) { $loadPrevious.after(element); });

                    if (data.isFirstPage) {
                        //Delete the tear and load previous button because we've loaded everything
                        $loadPreviousLink.unbind('click');
                        $loadPrevious.remove();
                        $topSpinner.remove();
                    } else {
                        //Set up for the next load request
                        lastFromId = fromId;
                        fromId = data.previousPageStartId;
                    }
                }).always(function() {
                    $topSpinner.spinStop();
                    $loadPreviousLink.show();
                });
            }

            $loadPreviousLink.click(function(e) {
                e.preventDefault();
                loadPreviousActivities();
            });
        };

        PullRequestActivity.prototype.attachActivities = function(values, attach) {
            var self = this;

            var index = 0;
            var $newItems = $(_.map(values, function(activity) {
                var activityItem = stash.feature.pullRequest.activityListItem({
                    activity : activity,
                    pullRequest: self.pullRequest.toJSON(),
                    isNew: false
                });
                index++;
                return activityItem;
            }).join(''));

            var dataByActivityId = _.reduce(values, function(map, activityData) {
                map[activityData.id] = activityData;
                return map;
            }, {});

            // Must attach before rendering diffs because the diffs need to know their width.
            // We never use 'html' because we have loaded one "item" already, which is the general activity form.
            // Doing this ruins our ability to "reset" (if we every want to switch to another activity stream) after
            // we've already loaded items. We'll have to manually reset if we ever want that.
            attach($newItems);

            this.renderedDiffFileContents = this.renderedDiffFileContents.concat(
                PullRequestActivity.renderDiffs(
                    $newItems.filter('.diff-comment-activity'),
                    dataByActivityId,
                    new CommitRange({
                        pullRequest : this.pullRequest
                    })
                )
            );
        };

        PullRequestActivity.prototype.attachNewContent = function(data, attachmentMethod) {
            var self = this;

            var $commentContainer = $('#pull-request-activity > li.comment-form-container');
            var isPermalinked = !_.isUndefined(this.fromId) && !_.isUndefined(this.fromType);
            var isCommentPermalink = this.fromType === 'comment';
            var isFirstAttach = $commentContainer.siblings().length === 0;

            if (isPermalinked && isFirstAttach) {
                this.decorateForFocus(data);
            }

            if (isPermalinked && !data.isFirstPage && isFirstAttach) {
                //Only show the tear and load-previous button if we are in fact
                //permalinking, we are not permalinked to the very first activity
                //and this is the very first page of results we are showing
                this.onAttachFirstPermalinkPage(data, attachmentMethod);
            }

            this.attachActivities(data.values,
                function(element) { self.attachContent(attachmentMethod, element); });    //attach fn

            this._$spinner.spinStop();
            if (data.isLastPage) {
                this._$spinner.remove();
            }

            if (isPermalinked && isFirstAttach) {
                var $focused = isCommentPermalink ? $('.comment.focused') : $('.activity-item.focused');
                var offset = $focused.offset();
                $('html, body').animate({
                    scrollTop: offset.top - document.documentElement.clientHeight / 4
                }, 500);
            }
        };

        PullRequestActivity.renderDiffForComment = function($container, data, commitRange) {
            var fileContent = new FileContent($container);
            var fileChange = FileChange.fromDiff(data.diff, commitRange);
            fileContent.init(fileChange, undefined, undefined, {
                commentMode : FileContent.commentMode.REPLY_ONLY,
                lineComments : [ data.comment ],

                contentType : FileContent.contentType.DIFF,
                changeTypeLozenge : false, //TODO maybe we can add this later? Don't have the data now though.
                changeModeLozenge : false,
                fileIcon: true,
                breadcrumbs : true,
                scrollPaneSelector : 'self',
                pullRequestDiffLink: true,
                // all attributes are returned as arrays of length 1
                pullRequestDiffCurrent: data.diff.attributes.current[0],
                pullRequestDiffLinkUrl: navbuilder.currentPullRequest()
                                    .diff()
                                    .change(fileChange.getPath())
                                    .build()
            });

            return fileContent;
        };

        PullRequestActivity.renderDiffs = function($diffContainers, preloadData, commitRange) {
            var elAndDatas = [];

            // reads first
            _.each($diffContainers, function(el) {
                var activityId = el.getAttribute('data-activityid');
                elAndDatas.push({
                    el : el,
                    data : preloadData[activityId]
                });
            });

            //then writes
            return _.map(elAndDatas, function(elAndData) {
                return PullRequestActivity.renderDiffForComment(elAndData.el, elAndData.data, commitRange);
            });
        };

        PullRequestActivity.prototype.addApprovalActivity = function (approved, user, date) {
            var $firstItem = $("#pull-request-activity .comment-form-container");
            $(stash.feature.pullRequest.activityApprovalListItem({
                user: user,
                action: approved ? 'APPROVED' : 'UNAPPROVED',
                activityDate: date
            })).hide().insertAfter($firstItem).fadeIn('slow');
        };

        // TODO
        PullRequestActivity.prototype.handleErrors = $.noop;

        return PullRequestActivity;
    });
