define('page/pull-requests-create',[
    'jquery',
    'underscore',
    'eve',
    'util/ajax',
    'util/navbuilder',
    'util/text',
    'util/dirty-tracker',
    'model/page-state',
    'model/revision-reference',
    'widget/confirm-dialog',
    'widget/branch-selector-dialog',
    'widget/user-multiselect',
    'widget/mentionable-textarea',
    'widget/markup-preview',
    'feature/commits/commits-table',
    'exports'
], function(
    $,
    _,
    eve,
    ajax,
    nav,
    textUtil,
    dirtyTracker,
    pageState,
    RevisionRef,
    ConfirmDialog,
    BranchSelectorDialog,
    UserMultiSelect,
    MentionableTextarea,
    markupPreview,
    CommitsTable,
    exports
) {
    var branchSelectors = [],
        $commitsContainer = $("#create-pull-request-commits"),
        $loadMoreTrigger = $commitsContainer.find('.load-more-trigger'),
        commitsTable,
        commitTablePreviewPageSize = 10,
        commitTableFullPageSize = 25,
        handlesNewlinesInTextCorrectly = ($('<div>').text('\n').text().length === 1);

    function getChangesetUrl(revisionRef) {
        return nav.currentRepo().changeset(revisionRef.getId()).build();
    }

    function initBranchSelectors(contexts) {
        contexts = [].concat(contexts || []); //initBranchSelectors can accept either a single context or an array of contexts

        _.each(contexts, function(context){
            var branchSelector = {};

            branchSelector.$field = $("#" + context);
            branchSelector.$branchSelectorButton = $(branchSelector.$field).next('button');
            branchSelector.branchSelectorDialog = new BranchSelectorDialog(branchSelector.$branchSelectorButton, context + "Selector", { branchesOnly : true, disableLinks: true, context: context });
            branchSelector.$headChangesetSpinner = $("<div class='spinner'/>").insertAfter(branchSelector.$branchSelectorButton);

            branchSelectors[context] = branchSelector;
        });

        eve.on('stash.widget.branchselector.revisionRefChanged', function(revisionRef, context) {
            var branchSelector = branchSelectors[context];

            if (!branchSelector || branchSelector.$field.val() === revisionRef.getId()) {
                //selector hasn't changed, don't update anything.
                eve.stop();
                return;
            }

            updateBranchSelector(branchSelector, revisionRef);
            updateChangesetBadge(branchSelector, revisionRef);
            updateSubmitButtonDisabled();

            eve('stash.page.create-pull-request.'+ context +'.revisionRefChanged', null, revisionRef, context);
        });
    }

    function updateBranchSelector(branchSelector, revisionRef) {
        branchSelector.$field.val(revisionRef.getId());
        branchSelector.branchSelectorDialog.updateSelectedRevisionReference(revisionRef);
    }

    function updateChangesetBadge(branchSelector, revisionRef) {
        var $changesetBadge = branchSelector.$field.siblings('.changeset-badge-detailed').hide();

        branchSelector.$headChangesetSpinner.show().spin('small');
        ajax.rest({
            url: getChangesetUrl(revisionRef)
        }).done(function (changeset) {
                var $newChangesetBadge = $(stash.feature.changeset.changesetBadge.detailed({
                    changeset: changeset,
                    repository: pageState.getRepository()
                })).hide();

                $changesetBadge.replaceWith($newChangesetBadge);
                $newChangesetBadge.fadeIn();
            }).always(function() {
                branchSelector.$headChangesetSpinner.spinStop().hide();
        });
    }

    function updateSubmitButtonDisabled() {
        var sourceBranch = branchSelectors.sourceBranch.$field.val(),
            targetBranch = branchSelectors.targetBranch.$field.val(),
            $submitButton = $("#submit-form");

        if (!(sourceBranch && targetBranch) || (sourceBranch === targetBranch)) {
            $submitButton.attr("disabled", true).addClass("disabled");
        } else {
            $submitButton.removeAttr("disabled").removeClass("disabled");
        }
    }

    function initMetaDataGenerator(){
        var $title = $("#title"),
            $description = $("#description"),
            sourceBranchName = branchSelectors.sourceBranch.$field.val(),
            $commitTableRows = $('#create-pull-request-commits-table').find('tbody tr');

        eve.on('stash.page.create-pull-request.sourceBranch.revisionRefChanged', function(newRevisionRef){
            updateDefaultTitle(generateTitleFromBranchName(newRevisionRef));
        });

        // Clear out the generated description whenever either branch changes, we aren't guaranteed that the `stash.widget.commitsTable.contentAdded` event will fire
        // (same branch selected for source and target won't trigger an update on the commits table)
        eve.on('stash.page.create-pull-request.*.revisionRefChanged', clearDefaultDescription);

        eve.on('stash.widget.commitsTable.contentAdded', function(content){
            if (content && !$description.attr('data-dirty')) {
                var description = generateDescriptionFromCommitTableRows(content.values);

                if (content.start > 0) {
                    //This is not the first page of commits so append the current description to the generated one
                    description += $description.val();
                }

                var maxLength = 30000;
                var andMore = '\n\n...';
                updateDefaultDescription(description.length > maxLength ? description.substring(0, maxLength - andMore.length) + andMore : description);
            }
        });

        if(!$title.val() && sourceBranchName) {
            //Generate initial title
            updateDefaultTitle(generateTitleFromBranchName(sourceBranchName));
        }

        if(!$description.val() && $commitTableRows.length) {
            //Generate initial description
            updateDefaultDescription(generateDescriptionFromCommitTableRows($commitTableRows));
        }
    }

    function generateTitleFromBranchName(branchNameOrRevisionRef){
        return textUtil.convertBranchNameToSentence((branchNameOrRevisionRef instanceof RevisionRef) ? branchNameOrRevisionRef.getDisplayId() : branchNameOrRevisionRef);
    }

    function updateDefaultTitle(title) {
        var $title = $("#title");

        if(title && !$title.attr("data-dirty")){
            $title.val(title);
        }
    }

    function generateDescriptionFromCommitTableRows(tableRows){
        if (!handlesNewlinesInTextCorrectly) {
            return;
        }

        var description = '',
            $tableRows = $(tableRows).filter(':not(.merge)'),
            messageIterator = _.bind(getCommitMessageFromTableRow, null, $tableRows.length > 1); //If there's only one commit, don't nest it in a list

        if ($tableRows.length) {
            description = _.map($tableRows, messageIterator)
                                .reverse()              //oldest commits first.
                                .join('')
                                .replace(/\n+/g, '\n'); //Compress any multi-newline sequences into single newlines
        }

        return description;
    }

    function getCommitMessageFromTableRow(shouldNestInList, tableRow) {
        var message = $(tableRow).find('.message span').text();

        return (shouldNestInList ? convertMessageToListItem(message) : message);
    }

    function convertMessageToListItem(message) {
        // Turn the commit message into a list item.
        // Indent bullet lists (`*`,`+`,`-` and numeric `1.` style variants supported) so they are nested under the commit list.
        var bulletsRegex = /\n([ \t]*([\*\+\-]|\d+\.)\s)/g,
            bulletReplacement = '\n' + textUtil.indent('$1');

        return '\n* ' + message.replace(bulletsRegex, bulletReplacement) ;
    }

    function updateDefaultDescription(description) {
        var $description = $("#description");

        if (description && !$description.attr("data-dirty")) {
            markupPreview.hideIfVisible($description.closest('form'));
            $description.val(description).trigger("keyup"); //Simulate keyup to trigger expanding textarea
        }
    }

    function clearDefaultDescription() {
        //clear the description if it's an unmodified default
        var $description = $("#description");

        if (!$description.attr("data-dirty")) {
            $description.val('');
        }
    }

    function getCommitsUrlBuilder() {
        return nav.currentRepo().commits().withParams({
                until : branchSelectors.sourceBranch.$field.val(),
                since:  branchSelectors.targetBranch.$field.val()
            });
    }

    function initCommitsTable() {
        commitsTable = new CommitsTable(getCommitsUrlBuilder, {
            target: "#create-pull-request-commits-table",
            pageSize: commitTableFullPageSize,
            allFetchedMessageHtml: stash_i18n('stash.web.pull-request.create.allcommitsfetched', 'No more commits to merge'),
            noneFoundMessageHtml: stash_i18n('stash.web.pull-request.create.nocommitsfetched', 'No commits to merge')
        });

        if (branchesSelected() && !branchesAreEqual()) {
            commitsTable.init({suspended:true});
        }

        eve.on('stash.page.create-pull-request.*.revisionRefChanged', function() {
            if (branchesSelected()) {
                var $createForm = $commitsContainer.parents('form');
                if (branchesAreEqual()) {
                    $commitsContainer.toggleClass('hidden', true);
                    commitsTable.reset();
                    $createForm
                        .find(".aui-message")
                            .remove()
                            .end()
                        .prepend(widget.aui.message.error({contents: stash_i18n('stash.web.pull-request.create.branches.equal', 'You cannot merge a branch with itself')}));
                } else {
                    $createForm
                        .find(".aui-message")
                        .remove();
                    $loadMoreTrigger.toggleClass('hidden', true);
                    $commitsContainer.toggleClass('hidden', false);
                    commitsTable.setOptions({'pageSize' : commitTablePreviewPageSize});
                    commitsTable.update({suspended:true});
                    commitsTable.setOptions({'pageSize' : commitTableFullPageSize});
                }
            } else {
                $commitsContainer.toggleClass('hidden', true);
            }
        });

        $loadMoreTrigger.find('a').on('click', function(e){
            e.preventDefault();

            if (!commitsTable.resume()) {
                commitsTable.loadAfter();
            }

            $loadMoreTrigger.toggleClass('hidden', true);
        });

        eve.on('stash.widget.commitsTable.contentAdded', function(data){
            if(data.start === 0 && data.size && !data.isLastPage) {
                $loadMoreTrigger.toggleClass('hidden', false);
            } else {
                $loadMoreTrigger.toggleClass('hidden', true);
            }
        });
    }

    function branchesSelected() {
        return !!(branchSelectors.sourceBranch.$field.val() && branchSelectors.targetBranch.$field.val());
    }

    function branchesAreEqual() {
        //Strip refs/heads/ from the comparison to handle when one of the branches was supplied via the querystring
        return branchSelectors.sourceBranch.$field.val().replace(/^refs\/heads\//, '') === branchSelectors.targetBranch.$field.val().replace(/^refs\/heads\//, '');
    }

    function initPageState(repositoryJSON) {
        pageState.setRepository(repositoryJSON);
        pageState.setProject(repositoryJSON.project);
    }

    function initMentionableTextarea($container) {
        this.mentionableTextarea = new MentionableTextarea({
            selector: '.pull-request-description textarea',
            $container: $container
        });
    }

    function initDescriptionPreview() {
        var $container = $('.pull-request-description');
        var $form = $container.closest('form');

        markupPreview.bindTo($form);
        $form.submit(function() {
            markupPreview.hideIfVisible($form);
        });
    }

    function initDirtyTracking() {
        var $title = $("#title"),
            $description = $("#description");

        $title.val() && $title.attr('data-dirty', true);
        $description.val() && $description.attr('data-dirty', true);

        dirtyTracker.track({container: '.pull-request-details'});
    }

    exports.onReady = function(repositoryJSON, submittedData) {
        initPageState(repositoryJSON);
        initBranchSelectors(["sourceBranch","targetBranch"]);
        initDirtyTracking();
        initMetaDataGenerator();
        initMentionableTextarea($('.pull-request-details'));
        initCommitsTable();
        initDescriptionPreview();
        var currentUser = pageState.getCurrentUser();
        new UserMultiSelect($('#reviewers'), submittedData.reviewers, [currentUser && currentUser.name]);
    };
});
