define('util/deprecation', [], function() {

    var deprecationAllowed = !/(\?|&)deprecation\b/.test(location.search);

    /**
     * Returns a wrapped version of the function that logs a deprecation warning when the function is used.
     * @param {Function} fn the fn to wrap
     * @param {string} name the name of the fn to be displayed in the message.
     * @return {Function}
     */
    function deprecateFunctionExpression(fn, name) {
        var called = false;
        return function() {
            if (!called || !deprecationAllowed) {
                var fnName = name || fn.name;
                var message = "This function " + (fnName ? "(" + fnName + ") " : "") + "has been deprecated and will be removed in a future release.";
                if (!deprecationAllowed) {
                    throw new Error(message);
                } else {
                    called = true;
                    console.log(message + " Add ?deprecation to your current URL to debug this usage.");
                }
            }
            return fn.apply(this, arguments);
        };
    }

    /**
     * Wraps an object method with a deprecation warning
     * @param {Object} obj An obj whose method you want to deprecate
     * @param {string} prop The property name of the method to be deprecated
     */
    function deprecateObjectMethod(obj, prop) {
        obj[prop] = deprecateFunctionExpression(obj[prop], prop);
    }

    /**
     * Deprecate an object property or function expression
     *     deprecate(obj, propertyName);
     *     deprecate(fn, fnName);
     * @param {(Function|Object)} fnOrObject
     * @param {string} name
     * @return {(Function|undefined)}
     */
    function deprecate(fnOrObject, name) {
        if (typeof fnOrObject === 'function') {
            return deprecateFunctionExpression(fnOrObject, name);
        } else {
            return deprecateObjectMethod(fnOrObject, name);
        }
    }

    return deprecate;
});
