define('util/navbuilder', [
    'underscore',
    'jquery',
    'aui',
    'lib/jsuri',
    'util/deprecation',
    'model/page-state',
    'model/path',
    'exports'
], function (
    _,
    $,
    AJS,
    Uri,
    deprecate,
    pageState,
    Path,
    exports
) {
    // The basic idea of the nav builder is to support a chain of method calls that roughly map to the structure of the
    // structure and categorisation of parts of Stash.
    //
    // E.g. nav.project('foo').repo('bar').permissions() will return a builder that can build the url for the permissions page
    // of the repository 'bar' of the project 'foo' while just nav.project('foo') will return a builder that for a url to
    // the project page for project 'foo')
    //
    // At each point in the method chain, the returned object will support methods that return a builder
    // that maps to a concept at a lower level.
    // At any point, withParams() may be called to add query parameters.
    // At most points, build() may be called to build the URL to be used. There are a few places where no URL makes sense.
    //
    // ** See the unit tests for a description of the urls that can be produced by the nav builder **

    /**
     * Encapsulates a uri path and params that make up a query string.
     * This class is immutable - all mutating operations return a new instance.
     */
    function PathAndQuery(components, params, anchor) {
        this.components = (_.isString(components) ? [components] : components) || [];
        this.params = params || {};
        this.anchor = anchor || undefined;
    }

    PathAndQuery.prototype.buildRelNoContext = function() {
        var path = '/' + _.map(this.components,
            function (component) {
                return encodeURIComponent(component);
            }).join('/');

        var query = _.map(this.params,
            function (value, key) {
                var encodedValue = encodeURIComponent(value);
                return encodeURIComponent(key) + (encodedValue ? '=' + encodedValue : '');
            }).join("&");

        return path + (query ? '?' + query : '') + (this.anchor ? '#' + this.anchor : '');
    };

    PathAndQuery.prototype.buildRelative = function () {
        return AJS.contextPath() + this.buildRelNoContext();
    };

    PathAndQuery.prototype.buildAbsolute = function () {
        return location.protocol + "//" + location.hostname + (location.port ? ':' + location.port : '') + this.buildRelative();
    };

    PathAndQuery.prototype.toString = function() {
        return this.buildRelative();
    };

    /**
     * Adds query parameters. If a map (object) is supplied, its properties are added to the parameters.
     * If a single string is supplied, it is added as a query parameter with no value.
     * @returns a new PathAndQuery object with the updated query params
     */
    PathAndQuery.prototype.addParams = function (params) {
        var path = new PathAndQuery(this.components, _.extend({}, this.params));
        if (_.isString(params)) {
            path.params[params] = '';
        } else {
            if (params.hasOwnProperty("queryParams")) {
                path.params = _.extend(path.params, params.queryParams);
            } else if (!params.hasOwnProperty("urlMode")) {
                path.params = _.extend(path.params, params);
            }// todo - implement urlMode
        }
        return path;
    };

    /**
     * Sets the document hash. If a hash has been set previously, it is overwritten
     * @return a new PathAndQuery object with unchanged path and query string params, but with a new anchor
     */
    PathAndQuery.prototype.withFragment = function (anchor) {
        return new PathAndQuery(this.components, this.params, anchor);
    };

    /**
     * Pushes a new path component onto the list of path components.
     * @returns a new PathAndQuery object with the updated query params
     */
    PathAndQuery.prototype.pushComponents = function() {
        var path = new PathAndQuery(this.components.slice(0), this.params);
        _.each(_.toArray(arguments).slice(0), function (component) {
            if (component !== '') {
                path.components.push(component);
            }
        });
        return path;
    };

    /**
     * Returns a new builder for the current path with the properties supported by otherMethods.
     * Used after each call in the method chain to construct the next object in the chain and specify
     * exactly which calls are acceptable.
     * @returns a new builder
     */
    PathAndQuery.prototype.makeBuilder = function (otherMethods) {
        var path = this;
        return _.extend({
            _path: function() { return path; },
            build: function() { return path.buildRelative(); },
            buildAbsolute: function() { return path.buildAbsolute(); },
            parse : function() { return parse(this.build()); },
            withParams: function(params) {
                //return a new builder with the same methods as the current builder but with an augmented (new) path
                return path.addParams(params).makeBuilder(otherMethods);
            },
            toString : function() { return this.build(); }
        }, otherMethods);
    };

    // id/slug/key helpers

    // If the input is a string, it's the key/slug/id.
    // Otherwise, if it exists, assume it's a project/repo/PR
    // Otherwise, use this pageState's project/repo/PR.

    function getProjectKey(projectOrKey) {
        if (typeof(projectOrKey) === 'string') {
            return projectOrKey;
        } else if (!projectOrKey) {
            throw new Error(stash_i18n('stash.web.error.no.project', "No project key was provided."));
        }
        return projectOrKey.key;
    }
    function getCurrentProject() {
        if (pageState.getProject()) {
            return pageState.getProject();
        }
        throw new Error(stash_i18n('stash.web.error.no.project.context', "There is no project in this context."));
    }

    function getRepoSlug(repoOrSlug) {
        if (typeof(repoOrSlug) === 'string') {
            return repoOrSlug;
        } else if (!repoOrSlug) {
            throw new Error(stash_i18n('stash.web.error.no.repo', "No repository slug was provided."));
        }
        return repoOrSlug.slug;
    }

    function getCurrentRepository() {
        if (pageState.getRepository()) {
            return pageState.getRepository();
        }
        throw new Error(stash_i18n('stash.web.error.no.repo.context', "There is no repository in this context."));
    }

    function getPullRequestId(prOrId) {
        if (typeof prOrId in {"string":1, "number": 1}) {
            return prOrId;
        } else if (!prOrId) {
            throw new Error(stash_i18n('stash.web.error.no.pull-request.id', "No pull request ID was provided."));
        }
        return prOrId.getId();
    }

    function getCurrentPullRequest() {
        if (pageState.getPullRequest()) {
            return pageState.getPullRequest();
        }
        throw new Error(stash_i18n('stash.web.error.no.pull-request.context', "There is no pull request in this context."));
    }

    //----------------------------------------
    //Start of fluent interface/method chains
    //----------------------------------------

    //--- Methods at the root of the chain ---

    function login() {
        return new PathAndQuery('login').makeBuilder({ next : loginNext});
    }

    function tmp() {
        return new PathAndQuery('tmp').makeBuilder({
            avatars: tmpAvatars
        });
    }

    function admin() {
        return new PathAndQuery('admin').makeBuilder({
            permissions: permissions,
            users: users,
            groups: groups,
            licensing: licensing,
            mailServer: mailServer,
            db : adminDb
        });
    }

    function allProjects() {
        return new PathAndQuery('projects').makeBuilder();
    }

    function captcha() {
        //Add a changing query param to ensure all browsers reload the image when refreshing it - some don't respect the HTTP headers
        return new PathAndQuery('captcha').addParams({'ts' : new Date().getTime().toString()}).makeBuilder();
    }

    function project(projectOrKey) {
        return new PathAndQuery(['projects', getProjectKey(projectOrKey)]).makeBuilder({
            allRepos: allRepos,
            repo: repo,
            settings: projSettings,
            permissions: permissions,
            remove: projDelete,
            users: users,
            groups: groups,
            avatar: projAvatar
        });
    }
    function currentProject() {
        return project(getCurrentProject());
    }

    function createProject() {
        return new PathAndQuery('projects').addParams('create').makeBuilder();
    }

    function rest() {
        return new PathAndQuery(['rest', 'api', 'latest']).makeBuilder({
            project: restProj,
            currentProject : restCurrentProject,
            currentRepo: restCurrentRepo,
            currentPullRequest: restCurrentPullRequest,
            markup: restMarkup,
            profile: restProfile,
            users: restUsers,
            groups: restGroups
        });
    }

    //--- Methods further down the chain ---

    function loginNext(url) {
        return this._path().addParams({ next : url }).makeBuilder();
    }

    function tmpAvatars(){
        return this._path().pushComponents('avatars').makeBuilder();
    }

    function permissions() {
        return this._path().pushComponents('permissions').makeBuilder({
            permission: permission,
            users: users,
            groups: groups
        });
    }

    function permission(name) {
        return this._path().pushComponents(name).makeBuilder({
            users: usersPerm,
            groups: groupsPerm,
            all: allPerm
        });
    }

    function usersPerm() {
        return this._path().pushComponents('users').makeBuilder();
    }

    function groupsPerm() {
        return this._path().pushComponents('groups').makeBuilder();
    }

    function allPerm() {
        return this._path().pushComponents('all').makeBuilder();
    }

    function anonPerm() {
        return this._path().pushComponents('anon').makeBuilder();
    }

    function nonePerm() {
        return this._path().pushComponents('none').makeBuilder();
    }

    function allRepos() {
        return this._path().pushComponents('repos').makeBuilder();
    }

    function users() {
        return this._path().pushComponents('users').makeBuilder({
            create: createEntity,
            deleteUser: deleteEntity,
            view: viewEntity,
            filter: filterEntity,
            deleteSuccess: deleteSuccess,
            permissions: permissions,
            none: nonePerm
        });
    }

    function groups() {
        return this._path().pushComponents('groups').makeBuilder({
            create: createEntity,
            deleteGroup: deleteEntity,
            view: viewEntity,
            filter: filterEntity,
            deleteSuccess: deleteSuccess,
            permissions: permissions,
            none: nonePerm
        });
    }

    function licensing() {
        return this._path().pushComponents('license').makeBuilder({
            edit: editLicense
        });
    }

    function editLicense() {
        return this._path().addParams({edit: ''}).makeBuilder();
    }

    function mailServer() {
        return this._path().pushComponents('mail-server').makeBuilder();
    }

    function adminDb() {
        return this._path().pushComponents('db').makeBuilder();
    }

    function createEntity() {
        return this._path().addParams({create: ''}).makeBuilder();
    }

    function deleteEntity(name) { // delete is a reserved keyword
        return this._path().addParams({name: name}).makeBuilder();
    }

    function viewEntity(name) {
        return this._path().pushComponents('view').addParams({name: name}).makeBuilder();
    }

    function filterEntity(filterValue) {
        return this._path().addParams({filter: filterValue}).makeBuilder();
    }

    function deleteSuccess(name) {
        return this._path().addParams({deleted: name}).makeBuilder();
    }

    function createRepo() {
        return this._path().pushComponents('repos').addParams('create').makeBuilder();
    }

    function projDelete() {
        return this._path().makeBuilder();
    }

    function projSettings() {
        return this._path().pushComponents('settings').makeBuilder();
    }

    function projAvatar(size) {
        var builder = this._path().pushComponents('avatar.png');
        if (size) {
            builder = builder.addParams({s: size});
        }
        return builder.makeBuilder();
    }

    function repo(repoOrSlug) {
        return this._path().pushComponents('repos', getRepoSlug(repoOrSlug)).makeBuilder({
            browse: repoBrowse,
            diff : repoDiff,
            commits: repoCommits,
            changeset: repoChangeset,
            settings: repoSettings,
            hooks: repoHooks,
            clone: repoClone,
            allPullRequests: allPullRequests,
            createPullRequest: createPullRequest,
            pullRequest: pullRequest,
            build: function () {
                return this._path().pushComponents('browse').toString(); //the stem /projects/PROJ/repos is different to the path needed if build() is called
            }
        });
    }
    function currentRepo() {
        return currentProject().repo(pageState.getRepository());
    }

    // pull path components from an arguments object.  We all .path('a', 'b') and .path(['a', 'b']) both.
    function componentsFromArguments(args) {

        //accept multiple args or accept a single arg that's an array to support .path('a', 'b') and .path(['a', 'b'])
        if (args.length === 1) {
            if (args[0] && args[0].getComponents) { // accept a Path object
                return args[0].getComponents();
            } else if ($.isArray(args[0])) {
                return args[0];
            }
        }
        return _.toArray(args);
    }

    function repoBrowse() {
        return this._path().pushComponents('browse').makeBuilder({
            path: repoBrowsePath,
            at: repoBrowsePathAt
        });
    }

    function repoBrowsePath() {
        var path = this._path();
        return path.pushComponents.apply(path, componentsFromArguments(arguments)).makeBuilder({
            at: repoBrowsePathAt,
            until: repoBrowsePathUntil,
            raw : repoBrowsePathAtRaw
        });
    }


    function repoBrowsePathAt(then) {
        var builder = this._path();
        if (then) {
            builder = builder.addParams({ at: then });
        }
        return builder.makeBuilder({
            until : repoBrowsePathUntil,
            raw : repoBrowsePathAtRaw
        });
    }

    function repoBrowsePathUntil(then) {
        return this._path().addParams({until: then}).makeBuilder({
            at: repoBrowsePathAt,
            raw : repoBrowsePathAtRaw
        });
    }

    function repoBrowsePathAtRaw() {
        return this._path().addParams({raw: ''}).makeBuilder();
    }


    function repoDiff(fileChangeOrPath) {
        var builder = this._path(),
            path;

        // Duck-type as adding FileChange as a dependency on navbuilder causes too much recursion in dependency stack
        var isFileChange = (
            fileChangeOrPath.getCommitRange &&
                fileChangeOrPath.getPath &&
                fileChangeOrPath.getSrcPath
            );

        if (isFileChange) {
            var commitRange = fileChangeOrPath.getCommitRange();
            path = fileChangeOrPath.getPath();
            if (commitRange.getPullRequest()) {
                builder = builder.pushComponents('pull-requests', commitRange.getPullRequest().getId());
            } else {
                builder = builder.addParams($.extend({}, {
                    until: commitRange.getUntilRevision() && commitRange.getUntilRevision().getId()
                }));
                var since = commitRange.getSinceRevision() && commitRange.getSinceRevision().getId();
                if (since) {
                    builder = builder.addParams({
                        since: since
                    });
                }
            }
        } else {
            path = fileChangeOrPath;
        }

        builder = builder.pushComponents('diff'); //need to do this separately otherwise we don't have the correct context for the next apply invocation.
        builder = builder.pushComponents.apply(builder, path && componentsFromArguments([path]));

        if (isFileChange && fileChangeOrPath.getSrcPath()) {
            builder = builder.addParams($.extend({}, { srcPath: fileChangeOrPath.getSrcPath().toString() }));
        }

        return builder.makeBuilder({
            at: repoDiffAt
        });
    }

    function repoDiffAt(then) {
        return this._path().addParams({at: then}).makeBuilder();
    }

    function repoCommits() {
        return this._path().pushComponents('commits').makeBuilder();
    }

    function repoChangeset(ref) {
        // a changeset ref can be a branch with a '/' in it, which essentially makes it a path component.
        var path = this._path().pushComponents('commits').makeBuilder()._path();
        return path.pushComponents.apply(path, new Path(ref).getComponents()).makeBuilder();
    }

    function repoSettings() {
        return this._path().pushComponents('settings').makeBuilder();
    }

    function repoHooks() {
        return this._path().pushComponents('hooks').makeBuilder();
    }

    function repoClone(dcvsKind) {
        var path = this._path();
        return new PathAndQuery(['scm', path.components[1], path.components[3] + '.' + dcvsKind], path.params).makeBuilder();
    }

    //--- Pull Request Methods ---

    function allPullRequests() {
        return this._path().pushComponents('pull-requests').makeBuilder();
    }

    function createPullRequest() {
        return this._path().pushComponents('pull-requests').addParams('create').makeBuilder({
            source: pullRequestSource,
            target: pullRequestTarget
        });
    }

    function pullRequestSource(source) {
        return this._path().addParams({source: source}).makeBuilder({
            target : pullRequestTarget
        });
    }

    function pullRequestTarget(target) {
        return this._path().addParams({target: target}).makeBuilder();
    }

    function pullRequest(prOrId) {
        return this._path().pushComponents('pull-requests', getPullRequestId(prOrId)).makeBuilder({
            unwatch: pullRequestUnwatch,
            /**
             * @deprecated since 2.0. Use overview() instead.
             */
            activity: pullRequestActivity,
            overview: pullRequestOverview,
            diff: pullRequestDiff,
            commits: pullRequestCommits,
            build: function () {
                return this._path().pushComponents('overview').toString(); //Default to overview view
            }
        });
    }
    function currentPullRequest() {
        return currentRepo.call(this).pullRequest(getCurrentPullRequest());
    }

    var pullRequestActivity = deprecate(pullRequestOverview, 'navbuilder::activity()');

    function pullRequestOverview(){
        return this._path().pushComponents('overview').makeBuilder({
            comment: pullRequestComment
        });
    }

    function pullRequestUnwatch(){
        return this._path().pushComponents('unwatch').makeBuilder();
    }

    function pullRequestComment(commentId){
        return this._path().addParams({commentId: commentId}).makeBuilder();
    }

    function pullRequestDiff(){
        return this._path().pushComponents('diff').makeBuilder({
            change: function (diffChangePath) {
                return this._path().withFragment(diffChangePath).makeBuilder();
            }
        });
    }

    function pullRequestCommits(){
        return this._path().pushComponents('commits').makeBuilder();
    }

    function restProj(projectOrKey) {
        var key = getProjectKey(projectOrKey);
        return this._path().pushComponents('projects', key).makeBuilder({
            allRepos: restAllRepos,
            repo: restRepo,
            permissions: restProjPermissions
        });
    }
    function restCurrentProject() {
        return restProj.call(this, getCurrentProject());
    }

    function restAllRepos() {
        return this._path().pushComponents('repos').makeBuilder();
    }

    function restRepo(repoOrSlug) {
        var slug = getRepoSlug(repoOrSlug);
        return this._path().pushComponents('repos', slug).makeBuilder({
            tags: restRepoTags,
            branches: restRepoBranches,
            commits : restRepoAllCommits,
            changeset: restRepoChangeset,
            changes: routeChangesRequest,
            browse: restRepoBrowse,
            files: restRepoFind,
            pullRequest : restPullRequest,
            allPullRequests : restAllPullRequests
        });
    }
    function restCurrentRepo() {
        return restCurrentProject.call(this).repo(getCurrentRepository());
    }

    function routeChangesRequest (commitRange) {
        if (commitRange.getPullRequest()) {
            return _.bind(restPullRequest, this)(commitRange.getPullRequest()).changes();
        } else if (commitRange.getUntilRevision()) {
            return _.bind(restRepoChanges, this)(commitRange.getUntilRevision().getId(), commitRange.getSinceRevision() && commitRange.getSinceRevision().getId());
        } else {
            throw new Error("A valid commit-range is required to retrieve changes");
        }
    }


    function restPullRequest(prOrId) {
        var id = getPullRequestId(prOrId);
        return this._path().pushComponents('pull-requests', id).makeBuilder({
            comments : function() { return this._path().pushComponents('comments').makeBuilder(); },
            comment : restComment,
            activities : function() { return this._path().pushComponents('activities').makeBuilder(); },
            changes: restPullRequestChanges,
            approve: restApprove,
            watch: restWatch,
            merge: restMerge,
            reopen: restReopen,
            decline: restDecline
        });
    }
    function restCurrentPullRequest() {
        return restCurrentRepo.call(this).pullRequest(getCurrentPullRequest());
    }

    function restAllPullRequests() {
        return this._path().pushComponents('pull-requests').makeBuilder();
    }

    function restComment(id) {
        return this._path().pushComponents('comments', id).makeBuilder({
            comment : restComment
        });
    }

    function restPullRequestChanges(){
        return this._path().pushComponents('changes').makeBuilder();
    }

    function restApprove() {
        return this._path().pushComponents('approve').makeBuilder();
    }

    function restWatch() {
        return this._path().pushComponents('watch').makeBuilder();
    }

    function restMerge() {
        return this._path().pushComponents('merge').makeBuilder();
    }

    function restDecline() {
        return this._path().pushComponents('decline').makeBuilder();
    }

    function restReopen() {
        return this._path().pushComponents('reopen').makeBuilder();
    }

    function restRepoTags() {
        return this._path().pushComponents('tags').makeBuilder();
    }

    function restRepoBranches() {
        return this._path().pushComponents('branches').makeBuilder();
    }

    function restRepoAllCommits() {
        return this._path().pushComponents('commits').makeBuilder();
    }

    function restRepoChangeset(csid) {
        return this._path().pushComponents('commits', csid).makeBuilder();
    }

    function restRepoBrowse() {
        return this._path().pushComponents('browse').makeBuilder({
            path: repoBrowsePath,
            at: repoBrowsePathAt
        });
    }

    function restRepoFind() {
        return this._path().pushComponents('files').makeBuilder({
            all: restRepoAllFiles,
            path: restRepoFilesInPath,
            at: repoBrowsePathAt
        });
    }

    function restRepoAllFiles() {
        return this._path().addParams({limit: 100000}).makeBuilder({
            path: restRepoFilesInPath,
            at: repoBrowsePathAt
        });
    }

    function restRepoFilesInPath() {
        var path = this._path();
        return path.pushComponents.apply(path, componentsFromArguments(arguments)).makeBuilder({
            at: repoBrowsePathAt
        });
    }

    function restRepoChanges(untilId, sinceId) {
        var builder = this._path().pushComponents('changes');
        if (sinceId) {
            builder = builder.addParams({since: sinceId});
        }

        return builder.addParams({until: untilId}).makeBuilder();
    }

    function restProjPermissions() {
        return this._path().pushComponents('permissions').makeBuilder({
            projectRead: restProjReadPerms,
            projectWrite: restProjWritePerms
        });
    }

    function restProjReadPerms() {
        return this._path().pushComponents('project-read').makeBuilder({
            all: restAllProjPerms,
            anon: restAnonProReadPerms
        });
    }

    function restProjWritePerms() {
        return this._path().pushComponents('project-write').makeBuilder({
            all: restAllProjPerms
        });
    }

    function restAllProjPerms() {
        return this._path().pushComponents('all').makeBuilder({
            allow: restAllowProjPerms
        });
    }

    function restAnonProReadPerms() {
        return this._path().pushComponents('anon').makeBuilder({
            allow: restAllowProjPerms
        });
    }

    function restAllowProjPerms(allow) {
        return this._path().addParams({'allow': allow}).makeBuilder();
    }

    function restMarkup() {
        return this._path().pushComponents('markup').makeBuilder({
            preview: restMarkupPreview
        });
    }

    function restMarkupPreview() {
        return this._path().pushComponents('preview').makeBuilder();
    }

    function restProfile() {
        return this._path().pushComponents('profile').makeBuilder({
            recent: restProfileRecent
        });
    }

    function restProfileRecent() {
        return this._path().pushComponents('recent').makeBuilder({
            repos: restProfileRecentRepos
        });
    }

    function restProfileRecentRepos() {
        return this._path().pushComponents('repos').makeBuilder();
    }

    function restUsers() {
        return this._path().pushComponents('users').makeBuilder();
    }

    function restGroups() {
        return this._path().pushComponents('groups').makeBuilder();
    }

  // HACKY CODE CHECK: off
    var fallbackUrlPattern = /(default-avatar-)\d+(\.png)/;
    function avatarUrl(person, size) {
        return {
            build : function() {
                var uri = parse(person.avatarUrl);
                if (uri.getQueryParamValue("s")) {
                    uri.replaceQueryParam("s", size);
                }
                return uri.toString().replace(fallbackUrlPattern, "$1" + size + "$2");
            }
        };
    }
    // HACKY CODE CHECK: on

    /* Parse URIs */
    function parse(uri) {
        return new Uri(uri);
    }

    //Append trailing slash to url if there isn't one already.
    function appendSlashToUrl(url) {
        if (typeof url !== 'string') {
            return '/';
        } else {
            return url + ((url.charAt(url.length - 1) !== '/') ? '/' : '');
        }
    }

    function newBuilder(components, params) {
        return new PathAndQuery(components, params);
    }


    exports.login = login;
    exports.tmp = tmp;
    exports.captcha = captcha;
    exports.admin = admin;
    exports.project = project;
    exports.allProjects = allProjects;
    exports.createProject = createProject;
    exports.currentProject = currentProject;
    exports.currentRepo = currentRepo;
    exports.currentPullRequest = currentPullRequest;
    exports.rest = rest;
    exports.parse = parse;
    exports.appendSlashToUrl = appendSlashToUrl;
    exports.avatarUrl = avatarUrl;
    exports.newBuilder = newBuilder;
});
