define('widget/paged-table',
    ['jquery', 'aui', 'eve', 'util/ajax', 'widget/paged-scrollable', 'widget/error-dialog'],
    function ($, AJS, eve, ajax, PagedScrollable, ErrorDialog) {

         /**
          * An abstract widget used for tables that wish to handle scroll events and load new rows as the user nears the edge of the page.
          *
          * Support is provided for tables that start with some row content already - the initial count of rows is used as the
          * start of the first page of rows to request.
          *
          * It is assumed sub-classes will want to allow navigation of rows and provision is made for this however it is not
          * strictly necessary for sub-classes to implement.
          *
          * To extend PagedTable, you must implement:
          * this.buildUrl(start, limit) : this must build a URL to retrieve the next page of row data
          * this.handleNewRows(data, attachmentMethod) : given a RestPage object and an attachmentMethod ('prepend', 'append', or 'html'),
          * should add new rows to the table
          * this.handleErrors(errors) : should display something appropriate in response to the errors encountered when retrieving new row data
          *
          * You must also supply the following options for which there are no defaults:
          *  - tableMessageClass - the custom CSS class for the element used to display table messages
          *  - allFetchedMessageHtml - the message shown when all rows have been fetched
          *  - noneFoundMessageHtml - the message shown when there were no rows to display
          *
          * You may wish to supply the following options to override defaults:
          *  - ajaxDataType (defaults to 'json') - the data type of the ajax response
          */
         function PagedTable(options) {
            options = $.extend({}, PagedTable.defaults, options);
            PagedScrollable.call(this, null, options);
            this.$table = $(options.target);
            this.$spinner = $("<div class='spinner'/>").hide().insertAfter(this.$table);
            this.spinnerShowing = false;
            this.processedDomData = (this.$table.data('lastpage') === undefined);
            // Has the data supplied in the initial page load been processed.
            // If there was none supplied (data-lastpage is undefined), this is set to true to skip loading the initial data from the DOM later on.
        }

        PagedTable.defaults = {
            ajaxDataType : 'json',
            spinnerSize : 'large'
        };

        $.extend(PagedTable.prototype, PagedScrollable.prototype);

        PagedTable.prototype.init = function() {
            return PagedScrollable.prototype.init.apply(this, arguments).done(function() {
                if (this.shortcutsInitialised) {
                    self.focusInitialRow();
                }
            });
        };

        PagedTable.prototype.requestData = function (start, limit) {
            if (!this.processedDomData) {
                return $.Deferred().resolve(this.createDataFromJQuery(start, limit, this.$table)); // The initial page will already be in the DOM
            }
            var self = this;
            self.spinnerShowing = true;
            self.$spinner.show().spin(self.options.spinnerSize);
            return ajax.ajax({
                url: self.buildUrl(start, limit),
                dataType: self.options.ajaxDataType,
                statusCode: self.options.statusCode
            }).always(function () {
                    // Ensure the spinner gets removed after the content is inserted to avoid jumps when
                    setTimeout(function () {
                        self.spinnerShowing && self.$spinner.spinStop().hide();
                        self.spinnerShowing = false;
                    }, 0);
                });
        };

        PagedTable.prototype.reset = function () {
            PagedScrollable.prototype.reset.call(this);
            if (this.processedDomData) {
                this.$table.children("tbody").empty().end()
                    .siblings('.' + this.options.tableMessageClass).remove().end();
            }
            this.$table.hide();
        };

        PagedTable.prototype.update = function (options) {
            this.reset();
            return this.init(options);
        };

        PagedTable.prototype.createDataFromJQuery = function (start, limit, $table) {
            var $rows = $table.children("tbody").children("tr"),
                size = $rows.size(),
                isLastPage = Boolean($table.data("lastpage"));

            return {
                start: start,
                size: size,
                values: $rows,
                isLastPage: isLastPage
            };
        };

        PagedTable.prototype.attachNewContent = function (data, attachmentMethod) {
            if (data.size) {
                if (this.processedDomData) {
                    if (data && data.errors) {
                        this.handleErrors(data.errors);
                        return;
                    }
                    this.handleNewRows(data, attachmentMethod);
                }

                if (data.start > 0 && data.isLastPage) {
                    this.handleLastPage();
                }
            } else {
                if (data.isLastPage) {
                    //There's no data and it's the last page, therefore there are no pull requests.
                    this.handleNoData();
                }
            }
            this.processedDomData = true;
        };

        /** Implement if rows can be focused and keyboard shortcuts allow focus navigation */
        PagedTable.prototype.focusInitialRow = function () {

        };

        /** Override if keyboard shortcuts are required but be sure to call this method afterwards */
        PagedTable.prototype.initShortcuts = function () {
            this.shortcutsInitialised = true;
            this.focusInitialRow(); //initShortcuts is called after init, so we need to manually call this the first time.
        };

        PagedTable.prototype.resetShortcuts = function () {
            this.shortcutsInitialised = false;
        };

        PagedTable.prototype.handleLastPage = function () {
            $("<p class='no-more-results'/>").addClass(this.options.tableMessageClass).html(this.options.allFetchedMessageHtml).insertAfter(this.$table);
        };

        PagedTable.prototype.handleNoData = function () {
            $("<h3 class='no-results entity-empty'/>").addClass(this.options.tableMessageClass).html(this.options.noneFoundMessageHtml).insertAfter(this.$table.hide());
        };

        PagedTable.prototype.handleErrors = function (errors) {
            new ErrorDialog({
                panelContent: '<p>' + AJS.escapeHtml(errors[0].message) + '</p>'
            }).show();
        };

        PagedTable.prototype.buildUrl = function (start, limit) {
            throw new Error('buildUrl is abstract and must be implemented.');
        };

        PagedTable.prototype.handleNewRows = function (data, attachmentMethod) {
            throw new Error('handleNewRows is abstract and must be implemented.');
        };

        return PagedTable;
    });
