define('widget/pager', ['jquery', 'underscore', 'util/ajax', 'widget/pager-filter'], function($, _, ajax, pageFilter) {
    function Pager(pagerEl) {
        this.$el = $(pagerEl);
    }

    Pager.prototype.getTarget = function() {
        return this.$el.data('for');
    };

    Pager.prototype.init = function(targetTemplateFn, siblingPagers, usePjax) {
        this.targetTemplateFn = targetTemplateFn;
        this.siblingPagers = siblingPagers;

        if (usePjax) {
            this._initPjax();
        } else {
            this._initAjax();
        }
    };

    Pager.prototype._updateSiblings = function(page) {
        //after the table data is updated, update the sibling pagers too
        _.each(this.siblingPagers, function(pager) {
            pager.update(page.start, page.limit, page.filter, page.isLastPage);
        });
    };

    Pager.prototype._initPjax = function() {
        var self = this,
            newPage = null;
        this.$el.find('a').pjax({
            dataType : 'json',
            timeout: 3000,
            template : function(page) {
                if (/*is a page*/ page && page.values ) {
                    //needed so that I can use the json object in the success handler
                    newPage = page;

                    return self.targetTemplateFn(page);
                }
            },
            success: function(unusableHtmlString) {
                //after the table data is updated, update the sibling pagers too
                self._updateSiblings(newPage);
                newPage = null;
            }
        });
    };

    /**
     * For paging without push state.  Do a normal AJAX request and populate the target.
     */
    Pager.prototype._initAjax = function() {
        var self = this;
        this.$el.on('click', 'a', function() {
            ajax.rest({
                url : this.href
            }).done(function(page) {
                self.refresh(page);
            });
            return false;
        });
    };

    Pager.prototype.refresh = function (page) {
        $("#" + this.getTarget()).html(this.targetTemplateFn(page));
        this._updateSiblings(page);
    };

    Pager.prototype.update = function(start, limit, filter, isLastPage) {
        var $newPager = $(stash.widget.pager({
            start : start,
            limit : limit,
            filter: filter ? filter : '',
            isLastPage : isLastPage,
            id : this.$el.attr('id'),
            forId : this.$el.data('for'),
            pageUrl : window.location.pathname
        }));

        this.$el.replaceWith($newPager);
        /* pjax uses .live.  Live doesn't work when your context is undefined (as is the case with dynamically generated elements)
         * To get a real context, we recreate the jQuery object from the element it generated
         */
        this.$el = $($newPager[0]);
        this._initPjax();
    };


    Pager.onReady = function(templatesByTarget, options) {

        options = $.extend({}, Pager.defaults, options);

        // grab all pagers on the page and group them by the element they are paging through
        var pagersByTarget = {};
        $('.pager').each(function() {
            var pager = new Pager(this),
                target = pager.getTarget();

            if (!pagersByTarget[target]) {
                pagersByTarget[target] = [];
            }

            pagersByTarget[target].push(pager);
        });

        // initialize each pager with the template for its target element and any sibling pagers it'll need to update.
        _.each(pagersByTarget, function(siblingPagers, target) {
            var templateFn = templatesByTarget[target];
            if (templateFn) {
                _.each(siblingPagers, function(pager) {
                    pager.init(templateFn, siblingPagers, options.usePushState);
                });
            }
        });

        pageFilter.onReady(options.filter, pagersByTarget, templatesByTarget);
    };

    Pager.defaults = {
        usePushState : true,
        filter : '.page-filter'
    };

    return Pager;
});
