define('widget/permission-table', [
    'jquery',
    'underscore',
    'aui',
    'model/page-state',
    'util/ajax',
    'util/navbuilder',
    'util/error',
    'util/warn-before-unload',
    'exports'
], function(
    $,
    _,
    AJS,
    pageState,
    ajax,
    nav,
    errorUtil,
    warnBeforeUnload,
    exports
) {

    function makeDataSource(navBuilder, entityType) {

        function getNonMembers(params) {

            return ajax.rest({
                url : navBuilder[entityType + 's']().none().build(),
                type : 'GET',
                data : params
            });

        }

        function getMembers(params) {
            return ajax.rest({
                url : navBuilder[entityType + 's']().build(),
                type : 'GET',
                data : params
            });

        }

        function setPermission(permission, id) {
            return ajax.rest({
                url : navBuilder[entityType + 's']().withParams({permission: permission, name : id }).build(),
                type : 'PUT',
                statusCode : {
                    '400' : function() { return false; },
                    '403' : function() { return false; },
                    '404' : function() { return false; },
                    '409' : function() { return false; }
                }
            });
        }

        function removePermissions(id) {
            return ajax.rest({
                url : navBuilder[entityType + 's']().withParams({ name : id }).build(),
                type : 'DELETE',
                statusCode : {
                    '400' : function() { return false; },
                    '403' : function() { return false; },
                    '404' : function() { return false; },
                    '409' : function() { return false; }
                }
            });
        }

        return {
            getMembers : getMembers,
            getNonMembers : getNonMembers,
            setPermission : setPermission,
            removePermissions : removePermissions
        };
    }





    function PermissionTable(container, permissionsNavBuilder, permissionsOrderedList, userOrGroup, currentUserHighestPerm, opts) {
        this._entityType = userOrGroup;

        if (this._entityType === 'user') {
            this._text = PermissionTable.userText;
            this._rowTemplate = stash.widget.userPermissionRow;
            this._isEntityCurrentUser = function(entity) {
                return pageState.getCurrentUser() && entity.name === pageState.getCurrentUser().name;
            };
        } else {
            this._text = PermissionTable.groupText;
            this._rowTemplate = stash.widget.groupPermissionRow;
            this._isEntityCurrentUser = function() { return false; };
        }

        this._dataSource = makeDataSource(permissionsNavBuilder, this._entityType);

        this._permissions = _.pluck(permissionsOrderedList, 'name');
        this._initialPermission = this._permissions[this._permissions.length - 1];
        this._permissionNames = _.pluck(permissionsOrderedList, 'i18nName');

        this._currentUserHighestPerm = currentUserHighestPerm;

        this._pageSize = 50;

        this._addedEntities = [];

        this.opts = opts || {};

        this._wireComponents(container);
        this._initAddButton();
        this._initRows();
        this._initLoadMoreButton();

        this.refresh();
    }

    PermissionTable.text = {
        addModal : {
            buttonDoingText: stash_i18n('stash.web.permission.table.filter.adding', "Adding..."),
            buttonDoneText: stash_i18n('stash.web.permission.table.filter.added', "Added"),
            closeButtonText: stash_i18n('stash.web.permission.table.filter.done', "Done"),
            addItemUnknownError : stash_i18n("stash.web.groups.picker.unexpected.error", "The element could not be added. Please consult the logs for more information.")
        },
        entityList : {
            buttonLoadMoreText: stash_i18n('stash.picker.load.more', "Load more")
        },
        pendingModification : stash_i18n('stash.web.pending.request', 'You have made permission changes that have not yet reached Stash. If you leave this page, those changes will be lost.'),
        cantGrantHigherPerms : function(permName) { return stash_i18n('stash.web.permission.table.cantgranthigher.tooltip', 'You cannot grant the {0} permission', permName); },
        inheritedPerm : stash_i18n('stash.web.permission.table.inheritedperm.tooltip', 'This permission is inherited from a higher permission')
    };

    PermissionTable.userText = $.extend(true, {
        inactiveEntity : function(name) {
            return stash_i18n('stash.web.permission.table.inactive.user',
                   '{0} is deleted or marked as inactive in a remote directory', name);
        },
        addModal : {
            title: stash_i18n('stash.web.permission.table.available.users', "Available Users"),
            filterPlaceholder: stash_i18n('stash.web.permission.table.users.search.placeholder', "Search users"),
            buttonText: stash_i18n('stash.web.permission.table.add.user', "Add User"),
            noItemsText: stash_i18n('stash.web.permission.table.no.users', "No users."),
            loadingItemsText: stash_i18n('stash.web.permission.table.loading.users', "Loading users...")
        },
        cantRevokeHigherPerms : function(permName) { return stash_i18n('stash.web.permission.table.cantrevokehigheruser.tooltip', 'You cannot change permissions for a {0} ', permName); },
        cantGrantSelfHigherPerms : function(permName) { return stash_i18n('stash.web.permission.table.cantgrantselfhigher.tooltip', 'You cannot grant yourself the {0} permission', permName); },
        implicitPerm : stash_i18n('stash.web.permission.table.basePerm.tooltip', 'To remove this permission, click the X to the right to remove the entire row.'),
        noMoreItemsText: stash_i18n("stash.web.permission.table.users.picker.none.added", "All users already have permissions set.")
    }, PermissionTable.text);

    PermissionTable.groupText =  $.extend(true, {
        inactiveEntity : function(name) {
            return stash_i18n('stash.web.permission.table.invisible.group', '{0} is deleted', name);
        },
        addModal : {
            title: stash_i18n('stash.web.permission.table.available.groups', "Available Groups"),
            filterPlaceholder: stash_i18n('stash.web.permission.table.group.search.placeholder', "Search groups"),
            buttonText: stash_i18n('stash.web.permission.table.add.group', "Add Group"),
            noItemsText:stash_i18n('stash.web.permission.table.no.groups',  "No groups."),
            loadingItemsText: stash_i18n('stash.web.permission.table.loading.groups', "Loading groups...")
        },
        cantRevokeHigherPerms : function(permName) { return stash_i18n('stash.web.permission.table.cantrevokehighergroup.tooltip', 'You cannot change permissions for a {0} group', permName); },
        implicitPerm : stash_i18n('stash.web.permission.table.basePerm.tooltip', 'To remove this permission remove the group row'),
        noMoreItemsText: stash_i18n("stash.web.permission.table.groups.picker.none.added", "All groups already have permissions set.")
    }, PermissionTable.text);

    PermissionTable.prototype._wireComponents = function(container) {
        this._$container = $(container);
        this._$addButton = this._$container.find('.add-button');
        this._$tbody = this._$container.find('tbody');
        this._$noResults = this._$container.find('.no-results-row');
        this._$loadMore = this._$container.find('.load-more');
        this._$spinner = $(this._$loadMore).parent().append('<div class="permissions-spinner" />').find('.permissions-spinner');
    };

    PermissionTable.prototype._getModalItem = function(restElem) {
        return {
            content: $(stash.widget.permissionModalItem({
                entity : restElem,
                isUser : this._entityType === 'user',
                inactiveText : this._text.inactiveEntity(restElem.name)
            })),
            id: restElem.name
        };
    };

    PermissionTable.prototype._initAddButton = function() {
        var self = this;

        var modalItemsCache = [];

        this._$addButton.bindModalFilteredPicker($.extend({
            buttonDoneClass: "green",
            panelClass: this._entityType + '-dialog-panel',
            pageSize: 25,
            retrieveItems: function(filter, offset, pageSize, callback) {
                self._dataSource.getNonMembers({
                    filter: filter,
                    start: offset,
                    limit: pageSize,
                    avatarSize: stash.widget.avatarSizeInPx({ size: 'xsmall' })
                }).done(function(data) {

                    modalItemsCache = modalItemsCache.concat(data.values);

                    var items = _.map(data.values, function(value) {
                        return self._getModalItem(value);
                    });

                    callback(items, !data.isLastPage, data.nextPageStart); //item indexes may be non contiguous - nextPageStart must be passed back in
                });
            },
            buttonPressed: function(id, doneCallback, errorCallback) {
                self._dataSource.setPermission(self._initialPermission, id)
                    .done(function() {

                        var entityForId = _.find(modalItemsCache, function(entity) {
                            return entity.name === id;
                        });
                        if (entityForId) {
                            self._addedEntities.push(entityForId);
                        }

                        if (doneCallback) {
                            doneCallback();
                        }
                    }).fail(function(xhr, textStatus, errorThrown, data) {
                        if (data && data.errors) {
                            errorCallback(data.errors);
                        } else {
                            errorCallback([{'message': self._text.addModal.addItemUnknownError}]);
                        }
                    });
            },
            onClose: function() {
                modalItemsCache = [];
                self.refresh();
            },
            isDisabled: function() {
                return self._$addButton.hasClass("disabled");
            },
            noMoreItemsText: this._text.noMoreItemsText,
            noItemsText: this._text.noItemsText
        }, this._text.addModal));
    };

    PermissionTable.prototype._updateErrors = function ($trigger, jqXhr) {
        var $errorHolder = $trigger.closest('tr').find('.permissions-item').parent();


        //immediately remove all previous errors in the current table.
        $errorHolder.closest('tbody').find('.error-message').remove();

        //if this request fails, add the new errors in.
        return jqXhr.fail(function(xhr, textStatus, errorThrown, data) {
            if (data && data.errors) {
                $.each(data.errors, function(i) {
                    $errorHolder.append($('<span class="error-message" />').text(data.errors[i].message));
                });
            }
        });
    };

    PermissionTable.prototype._spinWhilePending = function($trigger, promise) {

        if (promise && promise.state() === 'pending') {

            // disable all checkboxes for this row while request is pending
            $trigger.closest('tr').find(':checkbox').prop('disabled', true);

            var $spinner = $('<div class="spinner" />');
            $spinner.insertAfter($trigger);
            $trigger.addClass('hidden');

            $spinner.spin('small');

            promise.always(function() {
                $trigger.removeClass('hidden');
                $spinner.spinStop().remove();
            });
        }
    };

    PermissionTable.prototype._getAllCheckboxes = function($checkbox) {
        return $checkbox.parent().parent().find(':checkbox');
    };

    PermissionTable.prototype._getCheckboxStates = function($checkboxes) {
        return $.map($checkboxes, function(checkbox) {
            return { 'checked': $(checkbox).prop('checked'), 'disabled': $(checkbox).prop('disabled') };
        });
    };

    PermissionTable.prototype._restoreCheckboxStates = function($checkboxes, $states) {
        //Restore checked and disabled states to what they were previously
        $.each($states, function(index, state) {
            var $checkbox = $($checkboxes.get(index));

            $checkbox
                .prop('checked', state.checked)
                .prop('disabled', state.disabled)
                .toggleClass('disabled', state.disabled);
        });
    };

    PermissionTable.prototype._setChecked = function($newCheckbox, $oldCheckbox, $originalStates) {
        //Restore the checkbox states first, then update based on the selection
        this._restoreCheckboxStates(this._getAllCheckboxes($newCheckbox), $originalStates);

        //Uncheck the old one.
        $oldCheckbox.prop('checked', false).prop('title', '');

        //Check lower perms
        var lowerPerms = $newCheckbox.parent().nextAll().children(':checkbox')
            .prop('checked', true)
            .prop('disabled', true)
            .addClass('disabled')
            .prop('title', this._text.inheritedPerm);

        //Make the tooltip for the final perm indicate it is implicit rather than inherited
        //but only if it is the last perm left checked
        $newCheckbox.prop('title', lowerPerms.length ? '' : this._text.implicitPerm);

        //Check the new one.
        $newCheckbox
            .prop('checked', true)
            .prop('disabled', !lowerPerms.length) // disable it if it's the last one.
            .toggleClass('disabled', !lowerPerms.length);

        var name = $newCheckbox.closest('[data-entity-id]').attr('data-entity-id');
        _.each(this._addedEntities, function(entity) {
            if (entity.name === name) {
                entity.permission = $newCheckbox.val();
            }
        });
    };

    PermissionTable.prototype._initRows = function() {
        var self = this;

        this._$tbody.on('change', ':checkbox', function() {

            var $this = $(this);
            var $parent = $this.parent();
            var $next = $parent.next();

            var entityId = $parent.parent().attr('data-entity-id');
            var $oldCheckbox, $newCheckbox;
            var oldPerm, newPerm;

            if (this.checked) {
                // find the highest previously checked item.
                while($next.length && !$next.children(':checked').length) {
                    $next = $next.next();
                }

                $oldCheckbox = $next.children(':checkbox');
                $newCheckbox = $this;

                oldPerm = $oldCheckbox.prop('value');
                newPerm = this.value;
            } else {

                $oldCheckbox = $this;
                $newCheckbox = $next.children(':checkbox');

                oldPerm = this.value;
                newPerm = $newCheckbox.prop('value');
            }

            var requests = [];

            if (newPerm) {
                requests.push(self._dataSource.setPermission(newPerm, entityId));
            }

            //Capture states now prior to disabling them all
            var checkboxStates = self._getCheckboxStates(self._getAllCheckboxes($newCheckbox));

            var collectivePromise = $.when.apply($, requests);

            self._spinWhilePending($this, collectivePromise);
            self._updateErrors($this, collectivePromise);

            collectivePromise
                .done(function(){
                    //update permission level in UI
                    self._setChecked($newCheckbox, $oldCheckbox, checkboxStates);
                })
                .fail(function(xhr, textStatus, errorThrown, data) {
                    //revert to old permission level
                    self._setChecked($oldCheckbox, $newCheckbox, checkboxStates);
                });


            warnBeforeUnload(collectivePromise, self._text.pendingModification);

        }).on('click', '.delete-button', function(e) {
            // This is wrong, rather than telling the server we want to remove a specific permission,
            // we need a way to say "remove all permissions". Until the new rest endpoints are implemented though,
            // this will tell the server to remove the highest currently assigned permission (that we know about).
            // This will not work as intended if the user's permission level was changed in another window without
            // refreshing the current window (it will attempt to remove a permission level the user no longer has,
            // meaning the user is not removed)

            var $this = $(this),
                $parentRow = $this.parents('tr').first(),
                entityId = $parentRow.attr('data-entity-id'),
                $checkboxes = $parentRow.find(':checkbox'),
                checkboxStates = self._getCheckboxStates($checkboxes);


                var request = self._dataSource.removePermissions(entityId);

            self._spinWhilePending($this, request);
            self._updateErrors($this, request);

            request
                .done(function() {
                    self._numLoaded--;
                    // remove from addedEntities
                    self._addedEntities = _.filter(self._addedEntities, function(entity) {
                        if (entity.name === entityId && !entity.loaded) {
                            self._numLoaded++;
                        }
                        return entity.name !== entityId;
                    });

                    if (!$parentRow.siblings().length && !self._$loadMore.is(":visible")) {
                        var $noResultsInner = self._$noResults.find('div');

                        $noResultsInner.css('opacity', 0.0);
                        self._$noResults.show();
                        $noResultsInner.animate({'opacity': 1.0}, 500);
                    }

                    $parentRow.remove();
                })
                .fail(function(){
                    //revert to old permission level
                    self._restoreCheckboxStates($checkboxes, checkboxStates);
                });

            warnBeforeUnload(request, self._text.pendingModification);

            e.preventDefault();
            return false;
        });
    };

    PermissionTable.prototype._initLoadMoreButton = function(e) {
        var self = this;

        this._$loadMore.click(function(e) {
            self._loadItems();
            e.preventDefault();
        });
    };

    PermissionTable.prototype._renderRow = function(entity, grantedPermission, currentUserHighestPerm, isAdded) {
        var self = this;
        //Added users have no explicit granted perms - choose the last perm, the implicit perm
        var grantedPermissionIndex = _.isUndefined(grantedPermission) ? self._permissions.length - 1 : _.indexOf(this._permissions, grantedPermission);
        var currentUserHighestPermIndex = _.indexOf(this._permissions, currentUserHighestPerm);
        //This entity has a higher perm than the current user, their perms are unmodifiable
        var privilegedEntity = grantedPermissionIndex < currentUserHighestPermIndex;
        //The i18n permission name
        var grantedPermissionName = self._permissionNames[grantedPermissionIndex];
        return this._rowTemplate({
            entity : entity,
            entityPermissions :_.map(this._permissions, function(displayPermission, index, displayPermissions) {
                var permissionName = self._permissionNames[index];
                var isGranted = displayPermission === grantedPermission;
                //This privilege is higher than that of the current user
                var isPrivileged = index < currentUserHighestPermIndex;
                var isInherited = _.any(displayPermissions.slice(0, index), function(inheritingPermission) {
                    return inheritingPermission === grantedPermission;
                });
                return {
                    name : displayPermission,
                    granted : isGranted,
                    inherited : isInherited,
                    privileged : isPrivileged, //disabled if current user's highest perms not sufficient
                    tooltip : (function() {
                        if (privilegedEntity) {
                            return '';
                        } else if (isPrivileged && !isGranted) {
                            if (self._isEntityCurrentUser(entity)) {
                                return self._text.cantGrantSelfHigherPerms(permissionName);
                            } else {
                                return self._text.cantGrantHigherPerms(permissionName);
                            }
                        } else if (isInherited) {
                            return self._text.inheritedPerm;
                        } else if (index == self._permissions.length - 1) {
                            return self._text.implicitPerm;
                        } else {
                            return '';
                        }
                    }())
                };
            }),
            tooltip: privilegedEntity ? self._text.cantRevokeHigherPerms(grantedPermissionName) : '',
            showAddedLozenge : isAdded,
            showRemovePermsButton : isAdded || (grantedPermissionIndex >= currentUserHighestPermIndex),
            linkToEntities : pageState.getCurrentUser().isAdmin
        });
    };

    PermissionTable.prototype._loadItems = function() {
        var self = this;
        this._$spinner.spin('small');

        if (!this._numLoaded) {
            self._$tbody.empty();
            self._$noResults.hide();
        }

        // if we're starting fresh, render the newly added entities at the top.
        if (!this._numLoaded && this._addedEntities.length) {
            self._$tbody.append(
                _.chain(this._addedEntities)
                    .sortBy(function(entity) {
                        return entity.name;
                    })
                    .reduce(function(memo, entity) {
                        return memo + self._renderRow(entity, entity.permission, self._currentUserHighestPerm, true);
                    }, '')
                .value()
            );
        }

        self._dataSource.getMembers({
            start: self._numLoaded,
            limit: self._pageSize,
            avatarSize: stash.widget.avatarSizeInPx({ size: 'xsmall' })
        }).done(function(data) {
                if (self.opts.preProcessMembers) {
                    data = self.opts.preProcessMembers(data, self._entityType);
                }
                self._$spinner.spinStop();

                self._$loadMore.toggleClass('hidden disabled', data.isLastPage);

                if (data.values.length) {
                  self._$noResults.hide();
                } else if (!data.start && !self._addedEntities.length) {
                    // There are now no entities, show the noResults
                    self._$noResults.show();
                }

                self._numLoaded += data.values.length;
                // pluck out any of the added entities since we've moved them to the top.
                var addedEntityNames = _.invoke(_.pluck(self._addedEntities, 'name'), 'toLowerCase');

                var values = _.chain(data.values)
                                .filter(function(entity) {
                                    var addedEntityIndex = $.inArray(entity[self._entityType].name.toLowerCase(), addedEntityNames);
                                    if (addedEntityIndex !== -1) {
                                        self._addedEntities[addedEntityIndex].loaded = true;
                                    }
                                    return addedEntityIndex === -1;
                                }).value();

                if (values.length) {
                    self._$tbody.append(
                        _.reduce(values, function(memo, item) {
                            return memo + self._renderRow(item[self._entityType], item.permission, self._currentUserHighestPerm);
                        }, '')
                    );
                }

                self._$noResults.find("div").show();
        });

    };

    PermissionTable.prototype.refresh = function(hardRefresh) {
        // If it's a hard refresh, ignore the newly added entities
        if (hardRefresh) {
            this._addedEntities = [];
        }
        this._numLoaded = 0;
        _.each(this._addedEntities, function(entity) {
            entity.loaded = false;
        });
        this._loadItems();
    };

    /**
     * @param resNavBuilder           - the builder for the REST resource used for querying and updating permissions. Currently there are two
     *                                  compatible resources in Stash:
     *                                    "projects/{projectKey}/permissions" (for project permissions)
     *                                    "admin/permissions" (for global permissions)
     */
    exports.initialise = function(resNavBuilder, permissionsOrderedList, currentUserHighestPerm, opts) {
        return _.map(['user', 'group'], function(userOrGroup) {
            return new PermissionTable('#' + userOrGroup + '-permissions-table', resNavBuilder, permissionsOrderedList, userOrGroup, currentUserHighestPerm, opts);
        });
    };

});
